{"cells": [{"cell_type": "markdown", "metadata": {}, "source": ["# The Hydrogen Atom\n", "\n", "You will learn how to use *Psi4* by putting your hands on a simple\n", "example: The total energy of the hydrogen atom in Hartree-Fock theory.\n", "This is a tutorial - you are not only invited to type the commands that\n", "are being introduced, you are obliged to.\n", "\n", "## Electronic Structure Software\n", "\n", "*Ab initio* electronic structure software packages make it possible to\n", "calculate numerically a variety of properties of a given system, based\n", "only on physical constants and the system's Hamiltonian. The only\n", "approximations that need to be made are in the method and basis set that\n", "have to be chosen, in order to allow for a reasonable computational\n", "time. (The stronger your workstation, the more approximations you may\n", "drop, and the more elaborate your approach can be.) There are plenty of\n", "*ab initio* quantum chemical packages on the market; they differ in\n", "their capabilities, license policy and pricing. Widley used packages\n", "include GAMESS US, turbomole, DALTON, CP2K, CPMD and the Gaussian set of\n", "programs.\n", "\n", "*Psi4* is a free and open-source ab initio electronic structure program providing implementations of Hartree\u2013Fock, density functional theory, many-body perturbation theory, configuration interaction, density cumulant theory, symmetry-adapted perturbation theory, and coupled-cluster theory. Most of the methods are quite efficient, thanks to density fitting and multi-core parallelism. The program is a hybrid of C++ and Python, and calculations may be run with very simple text files or using the Python API, facilitating post-processing and complex workflows. [Reference: Smith DG, *et al.* PSI4 1.4: Open-source software for high-throughput quantum chemistry. *The Journal of chemical physics* (2020) DOI: https://doi.org/10.1063/5.0006002 ]\n", "\n", "The current version of *Psi4*  that you will be using on [noto.epfl](https://noto.epfl.ch/) is v1.5. \n", "\n", "**! Always remember to select the `Computational Chemistry` kernel to have psi4 and other useful packages available**."]}, {"cell_type": "markdown", "metadata": {}, "source": ["## Preliminary steps\n", "\n", "Before starting to work with *Psi4* we need to set up the environment, importing the required modules:"]}, {"cell_type": "code", "execution_count": 1, "metadata": {}, "outputs": [], "source": ["import psi4\n", "import pandas as pd\n", "import numpy as np\n", "import matplotlib.pyplot as plt\n", "import sys\n", "sys.path.append(\"..\")\n", "from helpers import *\n", "\n", "plt.style.use(['seaborn-poster', 'seaborn-ticks'])"]}, {"cell_type": "markdown", "metadata": {}, "source": ["then we can also set the maximum resources that can be used, specifying the available memory for *Psi4* to use with the `psi4.set_memory()` function, and the number of threads to use in SMP parallel computations with the `psi4.set_num_threads()` function"]}, {"cell_type": "code", "execution_count": 2, "metadata": {}, "outputs": [{"name": "stdout", "output_type": "stream", "text": ["\n", "  Memory set to   1.863 GiB by Python driver.\n", "  Threads set to 2 by Python driver.\n"]}], "source": ["psi4.set_memory('2 GB')\n", "psi4.set_num_threads(2)"]}, {"cell_type": "markdown", "metadata": {}, "source": ["## Writing An Input For *Psi4* and Invoking the Program\n", "\n", "The geomety of the $H$ atom can now be defined, passing it as a string into the `psi4.geometry()` function in either Z-matrix (see note below) or Cartesian format, where the triple-quote `\"\"\"string\"\"\"` syntax is used to allow the string to break over multiple lines. "]}, {"cell_type": "code", "execution_count": 3, "metadata": {}, "outputs": [], "source": ["h = psi4.geometry(\"\"\"\n", "0 2\n", "H 0.0 0.0 0.0\n", "\"\"\")"]}, {"cell_type": "markdown", "metadata": {}, "source": [":::{admonition} Atomic coordinates representation\n", ":class: dropdown\n", "\n", "* **Z-matrix**: each line gives the coordinates of a single atom in terms of its internal coordinates (atomic number, bond length, bond angle, and dihedral angle). As example, the Z-matrix for hydrogen peroxide is reported:\n", "\n", "```\n", "H\n", "O 1 0.9\n", "O 2 1.4 1 105.0\n", "H 3 0.9 2 105.0 1 120.0\n", "```\n", "\n", "The first line defines the first atom, a H in this case. The next line defines the second atom, an O, and specifies the internuclear distance (0.9\u00c5) with the first atom (1). The next line defines the third atom, another O, and specifies the internuclear distance (1.4\u00c5) with the second atom (2) and the angle (105.0\u00b0) with atom 2 and atom 1 (O-O-H angle). The last line defines the fourth atom, another H,  and specifies the internuclear distance (0.9\u00c5) with the third atom (3), the angle (105.0\u00b0) with atom 3 and atom 2 (H-O-O angle), and the dihedral angle  (120.0\u00b0) with atom 3, 2, and 1 (H-O-O-H angle).\\\n", "Note that if more atoms would be present, it would be needed to specify for each of them the value for  internuclear distance, angle and dihedral with other atoms previously defined.\n", "\n", "\n", "* **Cartesian coordinates**: each line gives the $x$, $y$, and $z$ coordinates of a single atom, with respect to the origin of the cartesian axes. The Cartesian coordinates for hydrogen peroxide result:\\\n", "\n", "```\n", "H 0.000 0.000 0.000\n", "O 0.900 0.000 0.000\n", "O 1.262 1.352 0.000\n", "H 1.742 1.465 0.753\n", "```\n", "Note that Z-matrices can be converted to Cartesian coordinates and back, as the structural information content is identical.\n", ":::"]}, {"cell_type": "markdown", "metadata": {}, "source": ["Where the first line contains the total charge of the system (0), followed by the spin multiplicity (2), and the following lines contain all the atoms of your system, one atom per line. In this case our moleule is composed of a single hydrogen atom, so we have a single line."]}, {"cell_type": "markdown", "metadata": {}, "source": ["You can visualize a 3D representation of your molecule using the `drawXYZ` function imported from the `helpers` we provide. In this case you will not see something particularly exciting, but it is always a good practice to have a visual look at your structures, to check that everythin is as expected. "]}, {"cell_type": "code", "execution_count": 5, "metadata": {}, "outputs": [{"data": {"application/3dmoljs_load.v0": "<div id=\"3dmolviewer_1665585409062793\"  style=\"position: relative; width: 400px; height: 400px\">\n        <p id=\"3dmolwarning_1665585409062793\" style=\"background-color:#ffcccc;color:black\">You appear to be running in JupyterLab (or JavaScript failed to load for some other reason).  You need to install the 3dmol extension: <br>\n        <tt>jupyter labextension install jupyterlab_3dmol</tt></p>\n        </div>\n<script>\n\nvar loadScriptAsync = function(uri){\n  return new Promise((resolve, reject) => {\n    var tag = document.createElement('script');\n    tag.src = uri;\n    tag.async = true;\n    tag.onload = () => {\n      resolve();\n    };\n  var firstScriptTag = document.getElementsByTagName('script')[0];\n  firstScriptTag.parentNode.insertBefore(tag, firstScriptTag);\n});\n};\n\nif(typeof $3Dmolpromise === 'undefined') {\n$3Dmolpromise = null;\n  $3Dmolpromise = loadScriptAsync('https://cdn.jsdelivr.net/npm/3dmol@latest/build/3Dmol-min.min.js');\n}\n\nvar viewer_1665585409062793 = null;\nvar warn = document.getElementById(\"3dmolwarning_1665585409062793\");\nif(warn) {\n    warn.parentNode.removeChild(warn);\n}\n$3Dmolpromise.then(function() {\nviewer_1665585409062793 = $3Dmol.createViewer($(\"#3dmolviewer_1665585409062793\"),{backgroundColor:\"white\"});\nviewer_1665585409062793.zoomTo();\n\tviewer_1665585409062793.addModel(\"1\\n\\n H    0.000000000000    0.000000000000    0.000000000000\\n\",\"xyz\");\n\tviewer_1665585409062793.setStyle({\"stick\": {}, \"sphere\": {\"scale\": 0.1}});\n\tviewer_1665585409062793.zoomTo();\nviewer_1665585409062793.render();\n});\n</script>", "text/html": ["<div id=\"3dmolviewer_1665585409062793\"  style=\"position: relative; width: 400px; height: 400px\">\n", "        <p id=\"3dmolwarning_1665585409062793\" style=\"background-color:#ffcccc;color:black\">You appear to be running in JupyterLab (or JavaScript failed to load for some other reason).  You need to install the 3dmol extension: <br>\n", "        <tt>jupyter labextension install jupyterlab_3dmol</tt></p>\n", "        </div>\n", "<script>\n", "\n", "var loadScriptAsync = function(uri){\n", "  return new Promise((resolve, reject) => {\n", "    var tag = document.createElement('script');\n", "    tag.src = uri;\n", "    tag.async = true;\n", "    tag.onload = () => {\n", "      resolve();\n", "    };\n", "  var firstScriptTag = document.getElementsByTagName('script')[0];\n", "  firstScriptTag.parentNode.insertBefore(tag, firstScriptTag);\n", "});\n", "};\n", "\n", "if(typeof $3Dmolpromise === 'undefined') {\n", "$3Dmolpromise = null;\n", "  $3Dmolpromise = loadScriptAsync('https://cdn.jsdelivr.net/npm/3dmol@latest/build/3Dmol-min.min.js');\n", "}\n", "\n", "var viewer_1665585409062793 = null;\n", "var warn = document.getElementById(\"3dmolwarning_1665585409062793\");\n", "if(warn) {\n", "    warn.parentNode.removeChild(warn);\n", "}\n", "$3Dmolpromise.then(function() {\n", "viewer_1665585409062793 = $3Dmol.createViewer($(\"#3dmolviewer_1665585409062793\"),{backgroundColor:\"white\"});\n", "viewer_1665585409062793.zoomTo();\n", "\tviewer_1665585409062793.addModel(\"1\\n\\n H    0.000000000000    0.000000000000    0.000000000000\\n\",\"xyz\");\n", "\tviewer_1665585409062793.setStyle({\"stick\": {}, \"sphere\": {\"scale\": 0.1}});\n", "\tviewer_1665585409062793.zoomTo();\n", "viewer_1665585409062793.render();\n", "});\n", "</script>"]}, "metadata": {}, "output_type": "display_data"}], "source": ["drawXYZ(h)"]}, {"cell_type": "markdown", "metadata": {}, "source": ["We can compute the energy of the system the function `psi4.energy()`, passing as first argument a string with the desired method and the basis set (`'METHOD/BASIS'`), and the target molecule as second argument. The ouput is the total electronic energy in Hartree. For this exercise we will use the unrestricted Hartree Fock method. You will see that in detail during next lectures and exercises and you will get familiar with it (for now it is sufficient to know we are using *a method*, with a particular basis set).\n", "\n", "Calling `psi4.energy()` will perform a so called **single point calculation** and will not change the geometry. During the calculation the program will perform a wavefunction optimization to find the lowest energy combination of wavefunction coefficients. \n", "\n", "We will save in separate log files the output of the calculations and we will store the energy value and print it in the next cell."]}, {"cell_type": "code", "execution_count": null, "metadata": {}, "outputs": [], "source": ["basissets = ['STO-3G', '6-31G', '6-311G'] # these are the basis sets we are going to use\n", "\n", "psi4.set_options({'reference':'UHF'}) # We are using Unrestriced Hartee Fock method\n", "\n", "for basis in basissets:\n", "    psi4.core.set_output_file(f'{basis}-h-output.log', False) # save in seperate log files\n", "    E = psi4.energy(f'hf/{basis}', molecule=h)  # we do the single point energy calculation once per basis set\n", "\n", "    print(basis, E)"]}, {"cell_type": "markdown", "metadata": {}, "source": ["To verify the number of basis sets used, navigate to the appropriate output log file located in your file directory (lefthand side of page). You can use the CTRL+F search function to look for the phrase \"Number of basis functions\" and read under the section called \"Primary Basis.\" "]}, {"cell_type": "markdown", "metadata": {}, "source": ["```{admonition} Exercise 7\n", ":class: exercise \n", "Include a table of the the calculated energies using the three different basis sets, specifying for each of them the number of basis functions used. Compare the energies with the analytical value for the H atom, given by the analytical expression: \n", "\n", "$$\n", " E = \\frac{1}{2} m_e c^2\\alpha^2,\n", "$$\n", "\n", "where $\\alpha$ is the fine structure constant. \n", "\n", "$$\\begin{align}\n", "m_e = 0.910953\\cdot10^{-30} kg\\\\\n", "c = 2.99792458\\cdot10^8 m s^{-1}\\\\\n", "\\alpha = 7.2973525376\\cdot10^{-3}\\\\\n", "N_A = 6.0221367\\cdot10^{23}mol^{-1}\n", "\\end{align}$$\n", "Pay attention to the units - use atomic units or kcal$\\cdot$mol$^{-1}$ throughout!\n", "```"]}, {"cell_type": "markdown", "metadata": {}, "source": ["```{admonition} Exercise 8\n", ":class: exercise \n", "What is the influence of the basis set size on the accuracy of the result? How do the split-valence bases compare to STO-3G?\n", "```"]}], "metadata": {"celltoolbar": "Tags", "kernelspec": {"display_name": "Computational Chemistry", "language": "python", "name": "compchem"}, "language_info": {"codemirror_mode": {"name": "ipython", "version": 3}, "file_extension": ".py", "mimetype": "text/x-python", "name": "python", "nbconvert_exporter": "python", "pygments_lexer": "ipython3", "version": "3.8.12"}}, "nbformat": 4, "nbformat_minor": 4}